@@ -25,8 +25,9 @@ if "bpy" in locals():
     ui = reload(ui)
     colors = reload(colors)
     tasks_queue = reload(tasks_queue)
+    rerequests = reload(rerequests)
 else:
-    from blenderkit import paths, append_link, utils, ui, colors, tasks_queue
+    from blenderkit import paths, append_link, utils, ui, colors, tasks_queue, rerequests
 
 import threading
 import time
@@ -79,7 +80,8 @@ def check_missing():
 
 def check_unused():
     '''find assets that have been deleted from scene but their library is still present.'''
-
+    #this is obviously broken. Blender should take care of the extra data automaticlaly
+    return;
     used_libs = []
     for ob in bpy.data.objects:
         if ob.instance_collection is not None and ob.instance_collection.library is not None:
@@ -95,7 +97,7 @@ def check_unused():
                 used_libs.append(ps.settings.instance_collection)
 
     for l in bpy.data.libraries:
-        if l not in used_libs:
+        if l not in used_libs and l.getn('asset_data'):
             print('attempt to remove this library: ', l.filepath)
             # have to unlink all groups, since the file is a 'user' even if the groups aren't used at all...
             for user_id in l.users_id:
@@ -189,7 +191,7 @@ def report_usages():
 
     for ob in asset_obs:
         asset_data = ob['asset_data']
-        abid = asset_data['asset_base_id']
+        abid = asset_data['assetBaseId']
 
         if assets.get(abid) is None:
             asset_usages[abid] = {'count': 1}
@@ -200,7 +202,7 @@ def report_usages():
     # brushes
     for b in bpy.data.brushes:
         if b.get('asset_data') != None:
-            abid = b['asset_data']['asset_base_id']
+            abid = b['asset_data']['assetBaseId']
             asset_usages[abid] = {'count': 1}
             assets[abid] = b['asset_data']
     # materials
@@ -210,7 +212,7 @@ def report_usages():
 
             if m is not None and m.get('asset_data') is not None:
 
-                abid = m['asset_data']['asset_base_id']
+                abid = m['asset_data']['assetBaseId']
                 if assets.get(abid) is None:
                     asset_usages[abid] = {'count': 1}
                     assets[abid] = m['asset_data']
@@ -237,7 +239,7 @@ def report_usages():
     scene['assets reported'] = assets_reported
 
     if new_assets_count == 0:
-        print('no new assets were added')
+        utils.p('no new assets were added')
         return;
     usage_report = {
         'scene': sid,
@@ -266,7 +268,6 @@ def report_usages():
         scene['assets rated'][k] = scene['assets rated'].get(k, False)
     thread = threading.Thread(target=utils.requests_post_thread, args=(url, usage_report, headers))
     thread.start()
-    # r = requests.post(url, headers=headers, json=usage_report)
     mt = time.time() - mt
     print('report generation:                ', mt)
 
@@ -281,49 +282,56 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
     # link the group we are interested in( there are more groups in File!!!! , have to get the correct one!)
     #
     scene = bpy.context.scene
-    scene['assets used'] = scene.get('assets used', {})
-    scene['assets used'][asset_data['asset_base_id']] = asset_data.copy()
-
-    scene['assets rated'] = scene.get('assets rated', {})
-
-    id = asset_data['asset_base_id']
-    scene['assets rated'][id] = scene['assets rated'].get(id, False)
 
     user_preferences = bpy.context.preferences.addons['blenderkit'].preferences
 
     if user_preferences.api_key == '':
         user_preferences.asset_counter += 1
 
-    if asset_data['asset_type'] == 'scene':
+    if asset_data['assetType'] == 'scene':
         scene = append_link.append_scene(file_names[0], link=False, fake_user=False)
         props = scene.blenderkit
         parent = scene
 
-    if asset_data['asset_type'] == 'model':
+    if asset_data['assetType'] == 'model':
         s = bpy.context.scene
         downloaders = kwargs.get('downloaders')
         s = bpy.context.scene
         sprops = s.blenderkit_models
-        # TODO this is here because combinations of linking objects or appending groups are rather not-userfull
-        if sprops.append_method == 'LINK_GROUP':
+        # TODO this is here because combinations of linking objects or appending groups are rather not-usefull
+        if sprops.append_method == 'LINK_COLLECTION':
             sprops.append_link = 'LINK'
             sprops.import_as = 'GROUP'
         else:
             sprops.append_link = 'APPEND'
             sprops.import_as = 'INDIVIDUAL'
 
-        # set consistency for objects allready in scene, otherwise this literally breaks blender :)
+        # copy for override
+        al = sprops.append_link
+        # set consistency for objects already in scene, otherwise this literally breaks blender :)
         ain = asset_in_scene(asset_data)
+
+        # override based on history
         if ain is not False:
             if ain == 'LINKED':
-                sprops.append_link = 'LINK'
-                sprops.import_as = 'GROUP'
+                al = 'LINK'
             else:
-                sprops.append_link = 'APPEND'
-                sprops.import_as = 'INDIVIDUAL'
+                al = 'APPEND'
+                if asset_data['assetType'] == 'model':
+                    source_parent = get_asset_in_scene(asset_data)
+                    parent, new_obs = duplicate_asset(source=source_parent, **kwargs)
+                    parent.location = kwargs['model_location']
+                    parent.rotation_euler = kwargs['model_rotation']
+                    # this is a case where asset is already in scene and should be duplicated instead.
+                    # there is a big chance that the duplication wouldn't work perfectly(hidden or unselectable objects)
+                    # so here we need to check and return if there was success
+                    # also, if it was successful, no other operations are needed , basically all asset data is already ready from the original asset
+                    if new_obs:
+                        bpy.ops.wm.undo_push_context(message='add %s to scene' % asset_data['name'])
+                        return
 
         # first get conditions for append link
-        link = sprops.append_link == 'LINK'
+        link = al == 'LINK'
         # then append link
         if downloaders:
             for downloader in downloaders:
@@ -337,19 +345,22 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
                                                        name=asset_data['name'])
                     return
 
-                if sprops.import_as == 'GROUP':
-                    parent, newobs = append_link.link_group(file_names[-1],
-                                                            location=downloader['location'],
-                                                            rotation=downloader['rotation'],
-                                                            link=link,
-                                                            name=asset_data['name'])
+                if link:
+                    parent, new_obs = append_link.link_collection(file_names[-1],
+                                                                 location=downloader['location'],
+                                                                 rotation=downloader['rotation'],
+                                                                 link=link,
+                                                                 name=asset_data['name'],
+                                                                 parent=kwargs.get('parent'))
 
                 else:
-                    parent, newobs = append_link.append_objects(file_names[-1],
+
+                    parent, new_obs = append_link.append_objects(file_names[-1],
                                                                 location=downloader['location'],
                                                                 rotation=downloader['rotation'],
                                                                 link=link,
-                                                                name=asset_data['name'])
+                                                                name=asset_data['name'],
+                                                                parent=kwargs.get('parent'))
                 if parent.type == 'EMPTY' and link:
                     bmin = asset_data['bbox_min']
                     bmax = asset_data['bbox_max']
@@ -357,17 +368,22 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
                     parent.empty_display_size = size_min
 
         elif kwargs.get('model_location') is not None:
-            if sprops.import_as == 'GROUP':
-                parent, newobs = append_link.link_group(file_names[-1],
-                                                        location=kwargs['model_location'],
-                                                        rotation=kwargs['model_rotation'],
-                                                        link=link,
-                                                        name=asset_data['name'])
+            if link:
+                parent, new_obs = append_link.link_collection(file_names[-1],
+                                                             location=kwargs['model_location'],
+                                                             rotation=kwargs['model_rotation'],
+                                                             link=link,
+                                                             name=asset_data['name'],
+                                                             parent=kwargs.get('parent'))
             else:
-                parent, newobs = append_link.append_objects(file_names[-1],
+                parent, new_obs = append_link.append_objects(file_names[-1],
                                                             location=kwargs['model_location'],
                                                             rotation=kwargs['model_rotation'],
-                                                            link=link)
+                                                            link=link,
+                                                            name=asset_data['name'],
+                                                            parent=kwargs.get('parent'))
+
+            # scale Empty for assets, so they don't clutter the scene.
             if parent.type == 'EMPTY' and link:
                 bmin = asset_data['bbox_min']
                 bmax = asset_data['bbox_max']
@@ -380,9 +396,9 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
             lib = group.library
             lib['asset_data'] = asset_data
 
-    elif asset_data['asset_type'] == 'brush':
+    elif asset_data['assetType'] == 'brush':
 
-        # TODO if allready in scene, should avoid reappending.
+        # TODO if already in scene, should avoid reappending.
         inscene = False
         for b in bpy.data.brushes:
 
@@ -401,9 +417,9 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
             shutil.copy(thumbpath, asset_thumb_path)
             brush.icon_filepath = asset_thumb_path
 
-        if bpy.context.sculpt_object:
+        if bpy.context.view_layer.objects.active.mode == 'SCULPT':
             bpy.context.tool_settings.sculpt.brush = brush
-        elif bpy.context.image_paint_object:  # could be just else, but for future possible more types...
+        elif bpy.context.view_layer.objects.active.mode == 'TEXTURE_PAINT':  # could be just else, but for future possible more types...
             bpy.context.tool_settings.image_paint.brush = brush
         # TODO set brush by by asset data(user can be downloading while switching modes.)
 
@@ -411,7 +427,7 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
         props = brush.blenderkit
         parent = brush
 
-    elif asset_data['asset_type'] == 'material':
+    elif asset_data['assetType'] == 'material':
         inscene = False
         for m in bpy.data.materials:
             if m.blenderkit.id == asset_data['id']:
@@ -429,17 +445,26 @@ def append_asset(asset_data, **kwargs):  # downloaders=[], location=None,
 
         parent = material
 
+    scene['assets used'] = scene.get('assets used', {})
+    scene['assets used'][asset_data['assetBaseId']] = asset_data.copy()
+
+    scene['assets rated'] = scene.get('assets rated', {})
+
+    id = asset_data['assetBaseId']
+    scene['assets rated'][id] = scene['assets rated'].get(id, False)
+
     parent['asset_data'] = asset_data  # TODO remove this??? should write to blenderkit Props?
+    bpy.ops.wm.undo_push_context(message='add %s to scene' % asset_data['name'])
     # moving reporting to on save.
     # report_use_success(asset_data['id'])
 
 
-@bpy.app.handlers.persistent
+# @bpy.app.handlers.persistent
 def timer_update():  # TODO might get moved to handle all blenderkit stuff, not to slow down.
     '''check for running and finished downloads and react. write progressbars too.'''
     global download_threads
     if len(download_threads) == 0:
-        return 1
+        return 1.0
     s = bpy.context.scene
     for threaddata in download_threads:
         t = threaddata[0]
@@ -465,7 +490,7 @@ def timer_update():  # TODO might get moved to handle all blenderkit stuff, not
             file_names = paths.get_download_filenames(asset_data)
             wm = bpy.context.window_manager
 
-            at = asset_data['asset_type']
+            at = asset_data['assetType']
             if ((bpy.context.mode == 'OBJECT' and (at == 'model' \
                                                    or at == 'material'))) \
                     or ((at == 'brush') \
@@ -493,15 +518,15 @@ def timer_update():  # TODO might get moved to handle all blenderkit stuff, not
                 else:
                     done = try_finished_append(asset_data, **tcom.passargs)
                     if not done:
-                        at = asset_data['asset_type']
+                        at = asset_data['assetType']
                         tcom.passargs['retry_counter'] = tcom.passargs.get('retry_counter', 0) + 1
                         if at in ('model', 'material'):
                             download(asset_data, **tcom.passargs)
-                        elif asset_data['asset_type'] == 'material':
+                        elif asset_data['assetType'] == 'material':
                             download(asset_data, **tcom.passargs)
-                        elif asset_data['asset_type'] == 'scene':
+                        elif asset_data['assetType'] == 'scene':
                             download(asset_data, **tcom.passargs)
-                        elif asset_data['asset_type'] == 'brush' or asset_data['asset_type'] == 'texture':
+                        elif asset_data['assetType'] == 'brush' or asset_data['assetType'] == 'texture':
                             download(asset_data, **tcom.passargs)
                     if bpy.context.scene['search results'] is not None and done:
                         for sres in bpy.context.scene['search results']:
@@ -509,7 +534,36 @@ def timer_update():  # TODO might get moved to handle all blenderkit stuff, not
                                 sres['downloaded'] = 100
 
                 utils.p('finished download thread')
-    return .2
+    return .5
+
+
+def download_file(asset_data):
+    # this is a simple non-threaded way to download files for background resolution genenration tool
+    file_name = paths.get_download_filenames(asset_data)[0]  # prefer global dir if possible.
+
+    if check_existing(asset_data):
+        # this sends the thread for processing, where another check should occur, since the file might be corrupted.
+        utils.p('not downloading, already in db')
+        return file_name
+    preferences = bpy.context.preferences.addons['blenderkit'].preferences
+    api_key = preferences.api_key
+
+    with open(file_name, "wb") as f:
+        print("Downloading %s" % file_name)
+        headers = utils.get_headers(api_key)
+
+        response = requests.get(asset_data['url'], stream=True)
+        total_length = response.headers.get('Content-Length')
+
+        if total_length is None:  # no content length header
+            f.write(response.content)
+        else:
+            dl = 0
+            for data in response.iter_content(chunk_size=4096):
+                dl += len(data)
+                print(dl)
+                f.write(data)
+    return file_name
 
 
 class Downloader(threading.Thread):
@@ -536,16 +590,22 @@ class Downloader(threading.Thread):
         api_key = self.api_key
 
         # TODO get real link here...
-        get_download_url(asset_data, scene_id, api_key, tcom=tcom)
+        has_url = get_download_url(asset_data, scene_id, api_key, tcom=tcom)
+
+        if not has_url:
+            tasks_queue.add_task(
+                (ui.add_report, ('Failed to obtain download URL for %s.' % asset_data['name'], 5, colors.RED)))
+            return;
         if tcom.error:
             return
-        # only now we can check if the file allready exists. This should have 2 levels, for materials and for brushes
+        # only now we can check if the file already exists. This should have 2 levels, for materials and for brushes
         # different than for the non free content. delete is here when called after failed append tries.
         if check_existing(asset_data) and not tcom.passargs.get('delete'):
             # this sends the thread for processing, where another check should occur, since the file might be corrupted.
             tcom.downloaded = 100
             utils.p('not downloading, trying to append again')
             return;
+
         file_name = paths.get_download_filenames(asset_data)[0]  # prefer global dir if possible.
         # for k in asset_data:
         #    print(asset_data[k])
@@ -565,14 +625,14 @@ class Downloader(threading.Thread):
             else:
                 tcom.file_size = int(total_length)
                 dl = 0
-                for data in response.iter_content(chunk_size=4096):
+                totdata = []
+                for data in response.iter_content(chunk_size=4096*32): #crashed here... why? investigate:
                     dl += len(data)
                     tcom.downloaded = dl
                     tcom.progress = int(100 * tcom.downloaded / tcom.file_size)
                     f.write(data)
                     if self.stopped():
                         utils.p('stopping download: ' + asset_data['name'])
-                        f.close()
                         os.remove(file_name)
                         return;
 
@@ -612,7 +672,6 @@ def download(asset_data, **kwargs):
         asset_data = copy.deepcopy(asset_data)
     else:
         asset_data = asset_data.to_dict()
-
     readthread = Downloader(asset_data, tcom, scene_id, api_key)
     readthread.start()
 
@@ -622,7 +681,7 @@ def download(asset_data, **kwargs):
 
 
 def check_downloading(asset_data, **kwargs):
-    ''' check if an asset is allready downloading, if yes, just make a progress bar with downloader object.'''
+    ''' check if an asset is already downloading, if yes, just make a progress bar with downloader object.'''
     global download_threads
 
     downloading = False
@@ -630,7 +689,7 @@ def check_downloading(asset_data, **kwargs):
     for p in download_threads:
         p_asset_data = p[1]
         if p_asset_data['id'] == asset_data['id']:
-            at = asset_data['asset_type']
+            at = asset_data['assetType']
             if at in ('model', 'material'):
                 downloader = {'location': kwargs['model_location'],
                               'rotation': kwargs['model_rotation']}
@@ -646,7 +705,7 @@ def check_existing(asset_data):
 
     file_names = paths.get_download_filenames(asset_data)
 
-    utils.p('check if file allready exists')
+    utils.p('check if file already exists')
     if len(file_names) == 2:
         # TODO this should check also for failed or running downloads.
         # If download is running, assign just the running thread. if download isn't running but the file is wrong size,
@@ -663,11 +722,11 @@ def check_existing(asset_data):
 
 
 def try_finished_append(asset_data, **kwargs):  # location=None, material_target=None):
-    ''' try to append asset, if not successfull delete source files.
+    ''' try to append asset, if not successfully delete source files.
      This means probably wrong download, so download should restart'''
     file_names = paths.get_download_filenames(asset_data)
     done = False
-    utils.p('try to append allready existing asset')
+    utils.p('try to append already existing asset')
     if len(file_names) > 0:
         if os.path.isfile(file_names[-1]):
             kwargs['name'] = asset_data['name']
@@ -679,29 +738,93 @@ def try_finished_append(asset_data, **kwargs):  # location=None, material_target
                 for f in file_names:
                     try:
                         os.remove(f)
-                    except:
-                        e = sys.exc_info()[0]
+                    except Exception as e:
+                        # e = sys.exc_info()[0]
                         print(e)
                         pass;
                 done = False
     return done
 
 
+def get_asset_in_scene(asset_data):
+    '''tries to find an appended copy of particular asset and duplicate it - so it doesn't have to be appended again.'''
+    scene = bpy.context.scene
+    for ob in bpy.context.scene.objects:
+        ad1 = ob.get('asset_data')
+        if not ad1:
+            continue
+        if ad1.get('assetBaseId') == asset_data['assetBaseId']:
+            return ob
+    return None
+
+
+def check_all_visible(obs):
+    '''checks all objects are visible, so they can be manipulated/copied.'''
+    for ob in obs:
+        if not ob.visible_get():
+            return False
+    return True
+
+
+def check_selectible(obs):
+    '''checks if all objects can be selected and selects them if possible.
+     this isn't only select_hide, but all possible combinations of collections e.t.c. so hard to check otherwise.'''
+    for ob in obs:
+        ob.select_set(True)
+        if not ob.select_get():
+            return False
+    return True
+
+
+def duplicate_asset(source, **kwargs):
+    '''Duplicate asset when it's already appended in the scene, so that blender's append doesn't create duplicated data.'''
+
+    # we need to save selection
+    sel = utils.selection_get()
+    bpy.ops.object.select_all(action='DESELECT')
+
+    # check visibility
+    obs = utils.get_hierarchy(source)
+    if not check_all_visible(obs):
+        return None
+    # check selectability and select in one run
+    if not check_selectible(obs):
+        return None
+
+    # duplicate the asset objects
+    bpy.ops.object.duplicate(linked=True)
+
+
+    nobs = bpy.context.selected_objects[:]
+    #get parent
+    for ob in nobs:
+        if ob.parent not in nobs:
+            parent = ob
+            break
+    # restore original selection
+    utils.selection_set(sel)
+    return parent , nobs
+
+
 def asset_in_scene(asset_data):
-    '''checks if the asset is allready in scene. If yes, modifies asset data so the asset can be reached again.'''
+    '''checks if the asset is already in scene. If yes, modifies asset data so the asset can be reached again.'''
     scene = bpy.context.scene
     au = scene.get('assets used', {})
 
+    id = asset_data['assetBaseId']
     if id in au.keys():
         ad = au[id]
         if ad.get('file_name') != None:
+
             asset_data['file_name'] = ad['file_name']
             asset_data['url'] = ad['url']
 
-            c = bpy.data.collections.get(ad['name'])
-            if c is not None:
-                if c.users > 0:
-                    return 'LINKED'
+            # browse all collections since linked collections can have same name.
+            for c in bpy.data.collections:
+                if c.name == ad['name']:
+                    # there can also be more linked collections with same name, we need to check id.
+                    if c.library and c.library.get('asset_data') and c.library['asset_data']['assetBaseId'] == id:
+                        return 'LINKED'
             return 'APPENDED'
     return False
 
@@ -724,18 +847,20 @@ def get_download_url(asset_data, scene_id, api_key, tcom=None):
         'scene_uuid': scene_id
     }
     r = None
+
     try:
-        r = requests.get(asset_data['download_url'], params=data, headers=headers)
+        r = rerequests.get(asset_data['download_url'], params=data, headers=headers)
     except Exception as e:
         print(e)
         if tcom is not None:
             tcom.error = True
-
     if r == None:
-        tcom.report = 'Connection Error'
-        tcom.error = True
+        if tcom is not None:
+            tcom.report = 'Connection Error'
+            tcom.error = True
         return 'Connection Error'
 
+
     if r.status_code < 400:
         data = r.json()
         url = data['filePath']
@@ -745,18 +870,17 @@ def get_download_url(asset_data, scene_id, api_key, tcom=None):
 
     if r.status_code == 403:
         r = 'You need Full plan to get this item.'
-        tcom.report = r
-        r1 = 'All materials and brushes are aviable for free. Only users registered to Standart plan can use all models.'
-        tasks_queue.add_task((ui.add_report, (r1, 5, colors.RED)))
-        tcom.error = True
+        # r1 = 'All materials and brushes are available for free. Only users registered to Standard plan can use all models.'
+        # tasks_queue.add_task((ui.add_report, (r1, 5, colors.RED)))
+        if tcom is not None:
+            tcom.report = r
+            tcom.error = True
 
-    if r.status_code == 401:
-        tcom.report = 'Invalid API key'
-        tcom.error = True
-        return 'Invalid API key'
     elif r.status_code >= 500:
-        tcom.report = 'Server error'
-        tcom.error = True
+        utils.p(r.text)
+        if tcom is not None:
+            tcom.report = 'Server error'
+            tcom.error = True
     return False
 
 
@@ -764,7 +888,7 @@ def start_download(asset_data, **kwargs):
     '''
     check if file isn't downloading or doesn't exist, then start new download
     '''
-    # first check if the asset is allready in scene. We can use that asset without checking with server
+    # first check if the asset is already in scene. We can use that asset without checking with server
     quota_ok = asset_in_scene(asset_data) is not False
 
     # otherwise, check on server
@@ -774,7 +898,7 @@ def start_download(asset_data, **kwargs):
     # is the asseet being currently downloaded?
     downloading = check_downloading(asset_data, **kwargs)
     if not downloading:
-        # check if there are files allready. This check happens 2x once here(for free assets),
+        # check if there are files already. This check happens 2x once here(for free assets),
         # once in thread(for non-free)
         fexists = check_existing(asset_data)
 
@@ -784,15 +908,15 @@ def start_download(asset_data, **kwargs):
         #     props = utils.get_search_props()
         #     props.report = str('asset ')
         if not done:
-            at = asset_data['asset_type']
+            at = asset_data['assetType']
             if at in ('model', 'material'):
                 downloader = {'location': kwargs['model_location'],
                               'rotation': kwargs['model_rotation']}
                 download(asset_data, downloaders=[downloader], **kwargs)
 
-            elif asset_data['asset_type'] == 'scene':
+            elif asset_data['assetType'] == 'scene':
                 download(asset_data, **kwargs)
-            elif asset_data['asset_type'] == 'brush' or asset_data['asset_type'] == 'texture':
+            elif asset_data['assetType'] == 'brush' or asset_data['assetType'] == 'texture':
                 download(asset_data)
 
 
@@ -807,10 +931,10 @@ asset_types = (
 
 
 class BlenderkitKillDownloadOperator(bpy.types.Operator):
-    """Kill a download."""
+    """Kill a download"""
     bl_idname = "scene.blenderkit_download_kill"
     bl_label = "BlenderKit Kill Asset Download"
-    bl_options = {'REGISTER', 'UNDO'}
+    bl_options = {'REGISTER', 'INTERNAL'}
 
     thread_index: IntProperty(name="Thread index", description='index of the thread to kill', default=-1)
 
@@ -823,10 +947,10 @@ class BlenderkitKillDownloadOperator(bpy.types.Operator):
 
 
 class BlenderkitDownloadOperator(bpy.types.Operator):
-    """Download and link asset to scene. Only link if asset allready available locally."""
+    """Download and link asset to scene. Only link if asset already available locally."""
     bl_idname = "scene.blenderkit_download"
     bl_label = "BlenderKit Asset Download"
-    bl_options = {'REGISTER', 'UNDO'}
+    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}
 
     asset_type: EnumProperty(
         name="Type",
@@ -836,14 +960,22 @@ class BlenderkitDownloadOperator(bpy.types.Operator):
     )
     asset_index: IntProperty(name="Asset Index", description='asset index in search results', default=-1)
 
-    target_object: StringProperty(
-        name="Material Target Object",
-        description="",
+    asset_base_id: StringProperty(
+        name="Asset base Id",
+        description="Asset base id, used instead of search result index.",
         default="")
+
+    target_object: StringProperty(
+        name="Target Object",
+        description="Material or object target for replacement",
+        default="")
+
     material_target_slot: IntProperty(name="Asset Index", description='asset index in search results', default=0)
     model_location: FloatVectorProperty(name='Asset Location', default=(0, 0, 0))
     model_rotation: FloatVectorProperty(name='Asset Rotation', default=(0, 0, 0))
 
+    replace: BoolProperty(name='Replace', description='replace selection with the asset', default=False)
+
     cast_parent: StringProperty(
         name="Particles Target Object",
         description="",
@@ -855,27 +987,62 @@ class BlenderkitDownloadOperator(bpy.types.Operator):
 
     def execute(self, context):
         s = bpy.context.scene
-        sr = s['search results']
 
-        asset_data = sr[self.asset_index].to_dict()  # TODO CHECK ALL OCCURANCES OF PASSING BLENDER ID PROPS TO THREADS!
+        if self.asset_index > -1:
+            # either get the data from search results
+            sr = s['search results']
+            asset_data = sr[
+                self.asset_index].to_dict()  # TODO CHECK ALL OCCURRENCES OF PASSING BLENDER ID PROPS TO THREADS!
+            asset_base_id = asset_data['assetBaseId']
+        else:
+            # or from the scene.
+            asset_base_id = self.asset_base_id
+
         au = s.get('assets used')
         if au == None:
             s['assets used'] = {}
-        if asset_data['asset_base_id'] in s.get('assets used'):
-            asset_data = s['assets used'][asset_data['asset_base_id']].to_dict()
-        atype = asset_data['asset_type']
-        if bpy.context.mode != 'OBJECT' and (
-                atype == 'model' or atype == 'material') and bpy.context.active_object is not None:
-            bpy.ops.object.mode_set(mode='OBJECT')
-        kwargs = {
-            'cast_parent': self.cast_parent,
-            'target_object': self.target_object,
-            'material_target_slot': self.material_target_slot,
-            'model_location': tuple(self.model_location),
-            'model_rotation': tuple(self.model_rotation)
-        }
+        if asset_base_id in s.get('assets used'):
+            # already used assets have already download link and especially file link.
+            asset_data = s['assets used'][asset_base_id].to_dict()
 
-        start_download(asset_data, **kwargs)
+        atype = asset_data['assetType']
+        if bpy.context.mode != 'OBJECT' and (
+                atype == 'model' or atype == 'material') and bpy.context.view_layer.objects.active is not None:
+            bpy.ops.object.mode_set(mode='OBJECT')
+
+        if self.replace:  # cleanup first, assign later.
+            obs = utils.get_selected_replace_adepts()
+            # print(obs)
+            for ob in obs:
+                print('replace attempt ', ob.name)
+                if self.asset_base_id != '':
+                    # this is for a case when replace is called from a panel, this makes the first of the objects not replacable.
+                    if ob.get('asset_data') is not None and ob['asset_data']['assetBaseId'] == self.asset_base_id:
+                        print('skipping this oneli')
+                        continue;
+
+                kwargs = {
+                    'cast_parent': self.cast_parent,
+                    'target_object': ob.name,
+                    'material_target_slot': ob.active_material_index,
+                    'model_location': tuple(ob.matrix_world.translation),
+                    'model_rotation': tuple(ob.matrix_world.to_euler()),
+                    'replace': False,
+                    'parent': ob.parent
+                }
+                utils.delete_hierarchy(ob)
+                start_download(asset_data, **kwargs)
+        else:
+            kwargs = {
+                'cast_parent': self.cast_parent,
+                'target_object': self.target_object,
+                'material_target_slot': self.material_target_slot,
+                'model_location': tuple(self.model_location),
+                'model_rotation': tuple(self.model_rotation),
+                'replace': False
+            }
+
+            start_download(asset_data, **kwargs)
         return {'FINISHED'}
 
 
@@ -884,7 +1051,9 @@ def register_download():
     bpy.utils.register_class(BlenderkitKillDownloadOperator)
     bpy.app.handlers.load_post.append(scene_load)
     bpy.app.handlers.save_pre.append(scene_save)
-    # bpy.app.timers.register(timer_update,  persistent = True)
+    user_preferences = bpy.context.preferences.addons['blenderkit'].preferences
+    if user_preferences.use_timers:
+        bpy.app.timers.register(timer_update)
 
 
 def unregister_download():
@@ -892,4 +1061,5 @@ def unregister_download():
     bpy.utils.unregister_class(BlenderkitKillDownloadOperator)
     bpy.app.handlers.load_post.remove(scene_load)
     bpy.app.handlers.save_pre.remove(scene_save)
-    # bpy.app.timers.unregister(timer_update)
+    if bpy.app.timers.is_registered(timer_update):
+        bpy.app.timers.unregister(timer_update)
