@@ -27,6 +27,7 @@ from io_scene_gltf2.io.com import gltf2_io
 from io_scene_gltf2.io.com import gltf2_io_constants
 from io_scene_gltf2.io.exp import gltf2_io_binary_data
 from . import gltf2_blender_export_keys
+from io_scene_gltf2.io.exp.gltf2_io_user_extensions import export_user_extensions
 
 
 @cached
@@ -37,11 +38,12 @@ def gather_animation_sampler(channels: typing.Tuple[bpy.types.FCurve],
                              bake_range_start,
                              bake_range_end,
                              action_name: str,
+                             driver_obj,
                              export_settings
                              ) -> gltf2_io.AnimationSampler:
 
     blender_object_if_armature = blender_object if blender_object.type == "ARMATURE" else None
-    if blender_object_if_armature is not None:
+    if blender_object_if_armature is not None and driver_obj is None:
         if bake_bone is None:
             pose_bone_if_armature = gltf2_blender_get.get_object_from_datapath(blender_object_if_armature,
                                                                                channels[0].data_path)
@@ -52,14 +54,15 @@ def gather_animation_sampler(channels: typing.Tuple[bpy.types.FCurve],
     non_keyed_values = __gather_non_keyed_values(channels, blender_object,
                                                  blender_object_if_armature, pose_bone_if_armature,
                                                  bake_channel,
+                                                 driver_obj,
                                                  export_settings)
 
 
-    return gltf2_io.AnimationSampler(
+    sampler = gltf2_io.AnimationSampler(
         extensions=__gather_extensions(channels, blender_object_if_armature, export_settings, bake_bone, bake_channel),
         extras=__gather_extras(channels, blender_object_if_armature, export_settings, bake_bone, bake_channel),
         input=__gather_input(channels, blender_object_if_armature, non_keyed_values,
-                             bake_bone, bake_channel, bake_range_start, bake_range_end, action_name, export_settings),
+                             bake_bone, bake_channel, bake_range_start, bake_range_end, action_name, driver_obj, export_settings),
         interpolation=__gather_interpolation(channels, blender_object_if_armature, export_settings, bake_bone, bake_channel),
         output=__gather_output(channels, blender_object.matrix_parent_inverse.copy().freeze(),
                                blender_object_if_armature,
@@ -69,79 +72,130 @@ def gather_animation_sampler(channels: typing.Tuple[bpy.types.FCurve],
                                bake_range_start,
                                bake_range_end,
                                action_name,
+                               driver_obj,
                                export_settings)
     )
 
+    export_user_extensions('gather_animation_sampler_hook',
+                            export_settings,
+                            sampler,
+                            channels,
+                            blender_object,
+                            bake_bone,
+                            bake_channel,
+                            bake_range_start,
+                            bake_range_end,
+                            action_name)
+
+    return sampler
+
 def __gather_non_keyed_values(channels: typing.Tuple[bpy.types.FCurve],
                               blender_object: bpy.types.Object,
                               blender_object_if_armature: typing.Optional[bpy.types.Object],
                               pose_bone_if_armature: typing.Optional[bpy.types.PoseBone],
                               bake_channel: typing.Union[str, None],
+                              driver_obj,
                               export_settings
                               ) ->  typing.Tuple[typing.Optional[float]]:
 
     non_keyed_values = []
 
-    if bake_channel is None:
-        target = channels[0].data_path.split('.')[-1]
-    else:
-        target = bake_channel
-    if target == "value":
-        return ()
+    obj = blender_object if driver_obj is None else driver_obj
 
-    indices = [c.array_index for c in channels]
-    indices.sort()
-    length = {
-        "delta_location": 3,
-        "delta_rotation_euler": 3,
-        "location": 3,
-        "rotation_axis_angle": 4,
-        "rotation_euler": 3,
-        "rotation_quaternion": 4,
-        "scale": 3,
-        "value": 1
-    }.get(target)
+    # Note: channels has some None items only for SK if some SK are not animated
+    if None not in channels:
+        # classic case for object TRS or bone TRS
+        # Or if all morph target are animated
 
-    if length is None:
-        # This is not a known target
-        return ()
+        if driver_obj is not None:
+            # driver of SK
+            return tuple([None] * len(channels))
 
-    for i in range(0, length):
-        if bake_channel is not None:
-            non_keyed_values.append({
-                "delta_location" : blender_object.delta_location,
-                "delta_rotation_euler" : blender_object.delta_rotation_euler,
-                "location" : blender_object.location,
-                "rotation_axis_angle" : blender_object.rotation_axis_angle,
-                "rotation_euler" : blender_object.rotation_euler,
-                "rotation_quaternion" : blender_object.rotation_quaternion,
-                "scale" : blender_object.scale
-            }[target][i])
-        elif i in indices:
-            non_keyed_values.append(None)
+        if bake_channel is None:
+            target = channels[0].data_path.split('.')[-1]
         else:
-            if blender_object_if_armature is None:
-                non_keyed_values.append({
-                    "delta_location" : blender_object.delta_location,
-                    "delta_rotation_euler" : blender_object.delta_rotation_euler,
-                    "location" : blender_object.location,
-                    "rotation_axis_angle" : blender_object.rotation_axis_angle,
-                    "rotation_euler" : blender_object.rotation_euler,
-                    "rotation_quaternion" : blender_object.rotation_quaternion,
-                    "scale" : blender_object.scale
-                }[target][i])
-            else:
-                 # TODO, this is not working if the action is not active (NLA case for example)
-                 trans, rot, scale = pose_bone_if_armature.matrix_basis.decompose()
-                 non_keyed_values.append({
-                    "location": trans,
-                    "rotation_axis_angle": rot,
-                    "rotation_euler": rot,
-                    "rotation_quaternion": rot,
-                    "scale": scale
-                    }[target][i])
+            target = bake_channel
+        if target == "value":
+            # All morph targets are animated
+            return tuple([None] * len(channels))
 
-    return tuple(non_keyed_values)
+        indices = [c.array_index for c in channels]
+        indices.sort()
+        length = {
+            "delta_location": 3,
+            "delta_rotation_euler": 3,
+            "location": 3,
+            "rotation_axis_angle": 4,
+            "rotation_euler": 3,
+            "rotation_quaternion": 4,
+            "scale": 3,
+            "value": len(channels)
+        }.get(target)
+
+        if length is None:
+            # This is not a known target
+            return ()
+
+        for i in range(0, length):
+            if bake_channel is not None:
+                non_keyed_values.append({
+                    "delta_location" : obj.delta_location,
+                    "delta_rotation_euler" : obj.delta_rotation_euler,
+                    "location" : obj.location,
+                    "rotation_axis_angle" : obj.rotation_axis_angle,
+                    "rotation_euler" : obj.rotation_euler,
+                    "rotation_quaternion" : obj.rotation_quaternion,
+                    "scale" : obj.scale
+                }[target][i])
+            elif i in indices:
+                non_keyed_values.append(None)
+            else:
+                if blender_object_if_armature is None:
+                    non_keyed_values.append({
+                        "delta_location" : obj.delta_location,
+                        "delta_rotation_euler" : obj.delta_rotation_euler,
+                        "location" : obj.location,
+                        "rotation_axis_angle" : obj.rotation_axis_angle,
+                        "rotation_euler" : obj.rotation_euler,
+                        "rotation_quaternion" : obj.rotation_quaternion,
+                        "scale" : obj.scale
+                    }[target][i])
+                else:
+                     # TODO, this is not working if the action is not active (NLA case for example)
+                     trans, rot, scale = pose_bone_if_armature.matrix_basis.decompose()
+                     non_keyed_values.append({
+                        "location": trans,
+                        "rotation_axis_angle": rot,
+                        "rotation_euler": rot,
+                        "rotation_quaternion": rot,
+                        "scale": scale
+                        }[target][i])
+
+        return tuple(non_keyed_values)
+
+    else:
+        # We are in case of morph target, where all targets are not animated
+        # So channels has some None items
+        first_channel = [c for c in channels if c is not None][0]
+        object_path = get_target_object_path(first_channel.data_path)
+        if object_path:
+            shapekeys_idx = {}
+            cpt_sk = 0
+            for sk in obj.data.shape_keys.key_blocks:
+                if sk == sk.relative_key:
+                    continue
+                if sk.mute is True:
+                    continue
+                shapekeys_idx[cpt_sk] = sk.name
+                cpt_sk += 1
+
+        for idx_c, channel in enumerate(channels):
+            if channel is None:
+                non_keyed_values.append(obj.data.shape_keys.key_blocks[shapekeys_idx[idx_c]].value)
+            else:
+                non_keyed_values.append(None)
+
+        return tuple(non_keyed_values)
 
 def __gather_extensions(channels: typing.Tuple[bpy.types.FCurve],
                         blender_object_if_armature: typing.Optional[bpy.types.Object],
@@ -170,6 +224,7 @@ def __gather_input(channels: typing.Tuple[bpy.types.FCurve],
                    bake_range_start,
                    bake_range_end,
                    action_name,
+                   driver_obj,
                    export_settings
                    ) -> gltf2_io.Accessor:
     """Gather the key time codes."""
@@ -181,6 +236,7 @@ def __gather_input(channels: typing.Tuple[bpy.types.FCurve],
                                                                                   bake_range_start,
                                                                                   bake_range_end,
                                                                                   action_name,
+                                                                                  driver_obj,
                                                                                   export_settings)
     times = [k.seconds for k in keyframes]
 
@@ -201,19 +257,33 @@ def __gather_interpolation(channels: typing.Tuple[bpy.types.FCurve],
                            bake_bone: typing.Union[str, None],
                            bake_channel: typing.Union[str, None]
                            ) -> str:
+
+    # Note: channels has some None items only for SK if some SK are not animated
+
     if gltf2_blender_gather_animation_sampler_keyframes.needs_baking(blender_object_if_armature,
                                                                      channels,
                                                                      export_settings):
         if bake_bone is not None:
+            # TODO: check if the bone was animated with CONSTANT
             return 'LINEAR'
         else:
-            max_keyframes = max([len(ch.keyframe_points) for ch in channels])
+            max_keyframes = max([len(ch.keyframe_points) for ch in channels if ch is not None])
             # If only single keyframe revert to STEP
-            return 'STEP' if max_keyframes < 2 else 'LINEAR'
+            if max_keyframes < 2:
+                return 'STEP'
 
-    blender_keyframe = channels[0].keyframe_points[0]
+            # If all keyframes are CONSTANT, we can use STEP.
+            if all(all(k.interpolation == 'CONSTANT' for k in c.keyframe_points) for c in channels if c is not None):
+                return 'STEP'
 
-    # Select the interpolation method. Any unsupported method will fallback to STEP
+            # Otherwise, sampled keyframes use LINEAR interpolation.
+            return 'LINEAR'
+
+    # Non-sampled keyframes implies that all keys are of the same type, and that the
+    # type is supported by glTF (because we checked in needs_baking).
+    blender_keyframe = [c for c in channels if c is not None][0].keyframe_points[0]
+
+    # Select the interpolation method.
     return {
         "BEZIER": "CUBICSPLINE",
         "LINEAR": "LINEAR",
@@ -231,6 +301,7 @@ def __gather_output(channels: typing.Tuple[bpy.types.FCurve],
                     bake_range_start,
                     bake_range_end,
                     action_name,
+                    driver_obj,
                     export_settings
                     ) -> gltf2_io.Accessor:
     """Gather the data of the keyframes."""
@@ -242,11 +313,12 @@ def __gather_output(channels: typing.Tuple[bpy.types.FCurve],
                                                                                   bake_range_start,
                                                                                   bake_range_end,
                                                                                   action_name,
+                                                                                  driver_obj,
                                                                                   export_settings)
     if bake_bone is not None:
         target_datapath = "pose.bones['" + bake_bone + "']." + bake_channel
     else:
-        target_datapath = channels[0].data_path
+        target_datapath = [c for c in channels if c is not None][0].data_path
 
     is_yup = export_settings[gltf2_blender_export_keys.YUP]
 
@@ -271,10 +343,12 @@ def __gather_output(channels: typing.Tuple[bpy.types.FCurve],
                          (0.0, 0.0, 1.0, 0.0),
                          (0.0, -1.0, 0.0, 0.0),
                          (0.0, 0.0, 0.0, 1.0)))
-                correction_matrix_local = gltf2_blender_math.multiply(axis_basis_change, bone.bone.matrix_local)
+                correction_matrix_local = axis_basis_change @ bone.bone.matrix_local
             else:
-                correction_matrix_local = gltf2_blender_math.multiply(
-                    bone.parent.bone.matrix_local.inverted(), bone.bone.matrix_local)
+                correction_matrix_local = (
+                    bone.parent.bone.matrix_local.inverted() @
+                    bone.bone.matrix_local
+                )
 
             transform = correction_matrix_local
         else:
@@ -338,4 +412,3 @@ def __gather_output(channels: typing.Tuple[bpy.types.FCurve],
         sparse=None,
         type=data_type
     )
-
